import uuid
from datetime import datetime
from typing import (
    Dict,
    ForwardRef,
    Generic,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from loguru import logger

from ..imports import *

ALL_EDGES = "allEdges"
SOURCE, TARGET, TYPE, EDGE_TYPE, LABEL, SEQUENCE = (
    "_source",
    "_target",
    "_type",
    "_type",
    "label",
    "sequence",
)


# typing utils for python 3.7
def get_args(type_annotation):
    return getattr(type_annotation, "__args__", tuple())


def get_origin(type_annotation):
    return getattr(type_annotation, "__origin__", None)


class Edge:
    """Edges makes a link between two `ItemBase` Items. You won't use this class a lot in practice, as edges are
    abstracted away for normal users. When items are retrieved from the database, the edges are parsed automatically.
    When you add an edge between to items within pymemri, you will often use `ItemBase.add_edge`"""

    def __init__(
        self, source, target, _type, label=None, sequence=None, created=False, reverse=True
    ):
        self.source = source
        self.target = target
        self._type = _type
        self.label = label
        self.sequence = sequence
        self.created = created
        self.reverse = reverse

    @classmethod
    def from_json(cls, json):
        from .schema import get_constructor

        # we only set the target here
        _type = json[EDGE_TYPE]
        json_target = json[TARGET]
        target_type = json_target["_type"]
        plugin_class = json_target.get("pluginClass", None)
        target_constructor = get_constructor(target_type, plugin_class)
        target = target_constructor.from_json(json_target)
        return cls(source=None, target=target, _type=_type)

    def __repr__(self):
        return f"{self.source} --{self._type}-> {self.target}"

    def update(self, api):
        if self.created:
            api.create_edges([self])

    def __eq__(self, other):
        return (
            self.source is other.source
            and self.target is other.target
            and self._type == other._type
            and self.reverse == other.reverse
            and self.created == other.created
            and self.label == other.label
        )

    def traverse(self, start):
        """We can traverse an edge starting from the source to the target or vice versa. In practice we often call
        item.some_edge_type, which calls item.traverse(edgetype), which in turn calls this function."""
        if start == self.source:
            return self.target
        elif start == self.target:
            return self.source
        else:
            raise ValueError


T = TypeVar("T")


def check_target_type(fn):
    """
    Decorator to perform type check the target type of the first argument, or list of arguments.
    """

    def _check_type_wrapper(self, arg):
        if isinstance(arg, Iterable):
            for item in arg:
                target_type_name = type(item.target).__name__
                if not (
                    target_type_name == self.target_type
                    or target_type_name
                    in [
                        t.__forward_arg__ if isinstance(t, ForwardRef) else t.__name__
                        for t in get_args(self.target_type)
                    ]
                ):
                    raise TypeError("Attempted to insert edge with invalid target type")
        elif isinstance(arg, Edge):
            target_type_name = type(arg.target).__name__
            if not (
                target_type_name == self.target_type
                or target_type_name
                in [
                    t.__forward_arg__ if isinstance(t, ForwardRef) else t.__name__
                    for t in get_args(self.target_type)
                ]
            ):
                raise TypeError("Attempted to insert edge with invalid target type")
        else:
            raise TypeError("Attempted to insert edge with invalid type")
        return fn(self, arg)

    return _check_type_wrapper


class EdgeList(list, Generic[T]):
    def __init__(
        self,
        name: str,
        target_type: Union[type, str, ForwardRef],
        data: List[Edge] = None,
    ) -> None:
        super().__init__()
        self.name = name

        if isinstance(target_type, type):
            target_type = target_type.__name__
        elif isinstance(target_type, ForwardRef):
            target_type = target_type.__forward_arg__
        self.target_type = target_type

        if data is not None:
            self.extend(data)

    @property
    def targets(self) -> List["Item"]:
        return [edge.target for edge in self]

    # Wrap all append, extend and add methods
    @check_target_type
    def append(self, item: Edge) -> None:
        return super().append(item)

    @check_target_type
    def extend(self, other: Iterable[Edge]) -> None:
        return super().extend(other)

    @check_target_type
    def __add__(self, other: Iterable[Edge]) -> "EdgeList":
        return super().__add__(other)

    @check_target_type
    def __iadd__(self, other: Iterable[Edge]) -> "EdgeList":
        return super().__iadd__(other)

    def __setitem__(self, i: int, item: Edge) -> None:
        if isinstance(item, Edge):
            target_type_name = type(item.target).__name__
            if not (
                target_type_name == self.target_type
                or target_type_name
                in [
                    t.__forward_arg__ if isinstance(t, ForwardRef) else t.__name__
                    for t in get_args(self.target_type)
                ]
            ):
                raise TypeError("Attempted to insert edge with invalid target type")
        else:
            raise TypeError("Attempted to insert edge with invalid type")
        return super().__setitem__(i, item)

    def insert(self, i: int, item: Edge) -> None:
        if isinstance(item, Edge):
            target_type_name = type(item.target).__name__
            if not (
                target_type_name == self.target_type
                or target_type_name
                in [
                    t.__forward_arg__ if isinstance(t, ForwardRef) else t.__name__
                    for t in get_args(self.target_type)
                ]
            ):
                raise TypeError("Attempted to insert edge with invalid target type")
        else:
            raise TypeError("Attempted to insert edge with invalid type")
        return super().insert(i, item)


class ItemBase:
    """Provides a base class for all items.
    All items in the schema inherit from this class, and it provides some
    basic functionality for consistency and to enable easier usage."""

    properties: List[str] = list()
    edges: List[str] = list()

    def __init__(self, id: str = None):
        self._date_local_modified = dict()
        self._in_pod: bool = False
        self._new_edges = list()
        self._original_properties = dict()

        self.id: Optional[str] = id

    def __setattr__(self, name, value):
        prev_val = getattr(self, name, None)
        super(ItemBase, self).__setattr__(name, value)

        if name in self.properties and value != prev_val:
            self._date_local_modified[name] = datetime.utcnow()
            if name not in self._original_properties:
                self._original_properties[name] = prev_val

    @property
    def _updated_properties(self):
        return set(self._original_properties.keys())

    def __getattribute__(self, name):
        val = object.__getattribute__(self, name)
        if name in object.__getattribute__(self, "edges"):
            if isinstance(val, Edge):
                return val.traverse(start=self)
            if isinstance(val, EdgeList):
                return [edge.traverse(start=self) for edge in val]
        return val

    def reset_local_sync_state(self):
        """
        reset_local_sync_state is called when self is created or updated (optionally via bulk) in the PodClient.
        """
        self._original_properties = dict()
        self._date_local_modified = dict()
        self._in_pod = True

    def add_edge(self, name, val):
        """Creates an edge of type name and makes it point to val"""
        if name not in self.edges:
            raise NameError(f"object {self} does not have edge with name {name}")

        existing = object.__getattribute__(self, name)
        edge = Edge(self, val, name, created=True)
        if edge not in existing:
            existing.append(edge)
            self._new_edges.append(edge)

    def is_expanded(self):
        """returns whether the node is expanded. An expanded node retrieved nodes that are
        *directly* connected to it
        from the pod, and stored their values via edges in the object."""
        return len(self.get_all_edges()) > 0

    def get_edges(self, name):
        return object.__getattribute__(self, name)

    def get_all_edges(self):
        return [e for attr in self.__dict__.values() if self.attr_is_edge(attr) for e in attr]

    def get_all_edge_names(self):
        return [k for k, v in self.__dict__.items() if self.attr_is_edge(v)]

    def get_property_names(self):
        return [k for k, v in self.__dict__.items() if not type(v) == list]

    @staticmethod
    def attr_is_edge(attr):
        return isinstance(attr, list) and len(attr) > 0 and isinstance(attr[0], Edge)

    def update(self, api, edges=True, create_if_not_exists=True, skip_nodes=False):

        if not self.exists(api):
            logger.debug(f"creating {self}")
            api.create(self)
        else:
            logger.debug(f"updating {self}")
            api.update_item(self)

        if edges:
            api.create_edges(self.get_all_edges())

    def exists(self, api):
        return api.exists(self.id) if self.id else None

    def create_id_if_not_exists(self):
        if self.id is None:
            self.id = uuid.uuid4().hex

    def store(self, client: "PodClient"):
        return client.add_to_store(self)

    def __repr__(self):
        id = self.id
        _type = self.__class__.__name__
        return f"{_type} (#{id})"

    @classmethod
    def from_data(cls, *args, **kwargs):
        edges = dict()
        new_kwargs = dict()
        for k, v in kwargs.items():
            if isinstance(v, ItemBase):
                edge = Edge(None, v, k)
                edges[k] = edge
                new_kwargs[k] = edge
            else:
                new_kwargs[k] = v

        res = cls(*args, **new_kwargs)

        for v in edges.values():
            v.source = res
        return res


class Item(ItemBase):
    """Item is the baseclass for all of the data classes."""

    properties = [
        "dateCreated",
        "dateModified",
        "dateServerModified",
        "deleted",
        "externalId",
        "itemDescription",
        "starred",
        "version",
        "id",
        "importJson",
        "pluginClass",
        "isMock",
    ]
    edges = ["label"]

    DATE_PROPERTIES = ["dateCreated", "dateModified", "dateServerModified"]

    def __init__(
        self,
        dateCreated: datetime = None,
        dateModified: datetime = None,
        dateServerModified: datetime = None,
        deleted: bool = None,
        externalId: str = None,
        itemDescription: str = None,
        starred: bool = None,
        version: str = None,
        id: str = None,
        importJson: str = None,
        pluginClass: str = None,
        isMock: bool = None,
        label: EdgeList["CategoricalPrediction"] = None,
    ):
        super().__init__(id)

        # Properties
        self.dateCreated: Optional[str] = dateCreated
        self.dateModified: Optional[str] = dateModified
        self.dateServerModified: Optional[str] = dateServerModified
        self.deleted: Optional[str] = deleted
        self.externalId: Optional[str] = externalId
        self.itemDescription: Optional[str] = itemDescription
        self.starred: Optional[str] = starred
        self.version: Optional[str] = version
        self.importJson: Optional[str] = importJson
        self.pluginClass: Optional[str] = pluginClass
        self.isMock: Optional[bool] = isMock

        # Edges
        self.label: EdgeList["CategoricalPrediction"] = EdgeList(
            "label", "CategoricalPrediction", label
        )

    @classmethod
    def parse_json(self, cls, json):
        property_kwargs = Item.parse_properties(cls, json)
        edge_kwargs = Item.parse_edges(cls, json)
        return {**property_kwargs, **edge_kwargs}

    @classmethod
    def parse_properties(self, cls, json):
        return {p: json.get(p, None) for p in cls.properties}

    @classmethod
    def parse_edges(self, cls, json):
        all_edges = json.get(ALL_EDGES, None)
        edge_kwargs = dict()
        reverse_edges = [f"~{e}" for e in cls.edges]
        if all_edges is not None:
            for edge_json in all_edges:
                edge = Edge.from_json(edge_json)
                if edge.type in self.edges + reverse_edges:
                    edge_name = self.remove_prefix(edge.type)
                    if edge_name in edge_kwargs:
                        edge_kwargs[edge_name] += [edge]
                    else:
                        edge_kwargs[edge_name] = [edge]
        return edge_kwargs

    @classmethod
    def get_property_types(cls, dates=False) -> Dict[str, type]:
        """
        Infer the property types of all properties in cls.
        Raises ValueError if type anotations for properties are missing in the cls init.
        """
        mro = cls.mro()
        property_types = dict()
        for basecls in reversed(mro[: mro.index(ItemBase)]):
            property_types.update(basecls.__init__.__annotations__)
        property_types = {k: v for k, v in property_types.items() if k in cls.properties}

        if not set(property_types.keys()) == set(cls.properties):
            raise ValueError(f"Item {cls.__name__} has missing property annotations.")

        res = dict()
        for k, v in property_types.items():
            if (
                k[:1] != "_"
                and k != "private"
                and not (isinstance(v, list))
                and v is not None
                and (not (dates == False and k in cls.DATE_PROPERTIES))
            ):
                res[k] = v
        return res

    @classmethod
    def get_edge_types(cls) -> List[Tuple[str, str, str]]:
        """
        Infer the types of all edges in cls as tuple (source_type, target_type)
        """
        mro = cls.mro()
        tgt_types = dict()
        for basecls in reversed(mro[: mro.index(ItemBase)]):
            tgt_types.update(basecls.__init__.__annotations__)
        tgt_types = {k: v for k, v in tgt_types.items() if k in cls.edges}

        res: list[tuple] = []
        for k, v in tgt_types.items():
            if hasattr(v, "__args__") and len(v.__args__):
                v = v.__args__[0]
                if get_origin(v) == Union:
                    for arg in get_args(v):
                        if isinstance(arg, ForwardRef):
                            res.append((k, cls.__name__, arg.__forward_arg__))
                        else:
                            res.append((k, cls.__name__, arg.__name__))
                    continue
                elif isinstance(v, type):
                    v = v.__name__
                elif isinstance(v, ForwardRef):
                    v = v.__forward_arg__
                else:
                    v = v
            else:
                v = "Any"
            res.append((k, cls.__name__, v))
        return res

    @classmethod
    def remove_prefix(s, prefix="~"):
        return s[1:] if s[0] == "`" else s

    @classmethod
    def from_json(cls, json):
        kwargs = Item.parse_json(cls, json)

        property_types = cls.get_property_types(dates=True)
        for k, v in kwargs.items():
            if v is not None and property_types[k] == datetime:
                # Datetime in pod is in milliseconds
                kwargs[k] = datetime.fromtimestamp(v / 1000.0)
        res = cls(**kwargs)
        for e in res.get_all_edges():
            e.source = res
        return res

    def _get_schema_type(self):
        for cls in self.__class__.mro():
            if cls.__name__ != "ItemBase":
                return cls.__name__

    def to_json(self, dates=True):
        res = dict()
        private = getattr(self, "private", [])
        for k, v in self.__dict__.items():
            if (
                k[:1] != "_"
                and k != "private"
                and k not in private
                and not (isinstance(v, list))
                and v is not None
                and (not (dates == False and k in self.DATE_PROPERTIES))
            ):
                if isinstance(v, datetime):
                    # Save datetimes in milliseconds
                    v = int(v.timestamp() * 1000)
                res[k] = v
        res["type"] = self._get_schema_type()
        return res



import re
import subprocess
import urllib
import zipfile
from pathlib import Path, PurePosixPath
from string import Template
from typing import Dict, List, Union

import giturlparse
import requests
from fastcore.script import Param, call_parse, store_true
from loguru import logger

import pymemri

TEMPLATE_URL = (
    "https://gitlab.memri.io/memri/plugin-templates/-/archive/dev/plugin-templates-dev.zip"
)
TEMPLATE_BASE_PATH = "plugin-templates-dev"


# If the owner of the repository is one of these groups, the CLI requires an additional `user` argument
GITLAB_GROUPS = ["memri", "plugins"]


def get_remote_url():
    path = Path(".")
    url = subprocess.getoutput(f"git config --get remote.origin.url")
    if not url:
        raise ValueError(
            f"You can only run this from a initialized gitlab repository, and '{path}' is not an initialized git repository"
        )
    parsed = giturlparse.parse(url)
    repo_url = parsed.url2https
    if repo_url.endswith(".git"):
        repo_url = repo_url[:-4]
    return repo_url


def infer_git_info(url):
    parsed = giturlparse.parse(url)
    return parsed.owner, parsed.repo


def download_file(url, fname=None):
    cert_path = Path(pymemri.__file__).parent / "cert" / "gitlab.memri.io.pem"
    r = requests.get(url, stream=True, verify=cert_path)
    fname = url.rsplit("/", 1)[1] if fname is None else fname
    with open(fname, "wb") as fd:
        for chunk in r.iter_content(chunk_size=128):
            fd.write(chunk)
    return fname


def str_to_identifier(s, lower=True):
    result = re.sub("\W|^(?=\d)", "_", s)
    if lower:
        result = result.lower()
    return result


def str_to_gitlab_identifier(s, lower=True):
    result = re.sub("\W|^(?=\d)", "-", s)
    if lower:
        result = result.lower()
    return result


def reponame_to_displayname(reponame: str) -> str:
    return re.sub("[-_]+", " ", reponame).title()


def download_plugin_template(
    template_name: str, url: str = TEMPLATE_URL, base_path: str = TEMPLATE_BASE_PATH
):
    # base_path = Path(base_path) / template_name if url is None else Path(url.rsplit("/", 1)[1].split(".")[0]) / template_name

    zip_path = download_file(url)
    base_path = Path(zip_path.split(".")[0]) / template_name

    with zipfile.ZipFile(zip_path, "r") as f:
        result = {name: f.read(name) for name in f.namelist() if base_path in Path(name).parents}

    if len(result) == 0:
        Path(zip_path).unlink()
        raise ValueError(f"Could not find template: {template_name}")

    result = {
        str(PurePosixPath(k).relative_to(PurePosixPath(base_path))): v.decode("utf-8")
        for k, v in result.items()
        if v
    }
    Path(zip_path).unlink()
    return result


def get_templates(url: str = TEMPLATE_URL) -> List[str]:
    zip_path = download_file(url)

    with zipfile.ZipFile(zip_path, "r") as f:
        files_split = [name.split("/") for name in f.namelist()]
        result = [fn[1] for fn in files_split if fn[-1] == "" and len(fn) == 3]
    return result


class TemplateFormatter:
    def __init__(
        self,
        template_dict: Dict[str, str],
        replace_dict: Dict[str, str],
        tgt_path: Union[str, Path],
        verbose: bool = False,
    ):
        self.template_dict = template_dict
        self.tgt_path = Path(tgt_path)
        self.replace_dict = replace_dict
        self.verbose = verbose

    def format_content(self, content):
        return Template(content).safe_substitute(self.replace_dict)

    def format_path(self, path):
        new_path = Template(path).safe_substitute(self.replace_dict)
        return self.tgt_path / new_path

    def format_file(self, filename, content):
        new_path = self.format_path(filename)
        new_content = self.format_content(content)
        new_path.parent.mkdir(exist_ok=True, parents=True)
        if self.verbose:
            logger.info(f"Formatting {filename} -> {new_path}")
        with open(new_path, "w", encoding="utf-8") as f:
            f.write(new_content)

    def get_files(self):
        return [self.format_path(filename) for filename in self.template_dict.keys()]

    def format(self):
        for filename, content in self.template_dict.items():
            self.format_file(filename, content)

    def print_filetree(self):
        previous_prefix = None
        res = "Created the following files"

        for path in sorted(
            [x.relative_to(self.tgt_path) for x in self.get_files()],
            key=lambda item: 100 * str(item).count("/"),
        ):
            n_slashes = str(path).count("/")
            new_prefix = path.parent
            if previous_prefix != new_prefix and str(new_prefix) != ".":
                res = f"{res}\n├── {new_prefix}"
            if n_slashes == 0:
                res = f"{res}\n├── {path}"
            elif n_slashes == 1:
                res = f"{res}\n│   ├── {path.name}"

            previous_prefix = new_prefix

        print(res.strip() + "\n")


def get_template_replace_dict(
    repo_url=None,
    user=None,
    plugin_name=None,
    package_name=None,
    description=None,
    install_requires=None,
    template_name=None,
):
    if repo_url is None:
        repo_url = get_remote_url()

    try:
        repo_owner, repo_name = infer_git_info(repo_url)
    except ValueError:
        url_inf, owner_inf, name_inf = None, None, None
        logger.error(
            "Could not infer git information from current directory, no initialized repository found."
        )

    if repo_url is None:
        repo_url = url_inf

    if user is None:
        if repo_owner in GITLAB_GROUPS:
            user = None
        else:
            user = repo_owner

    if plugin_name is None:
        if repo_name is None:
            plugin_name = None
        else:
            plugin_name = reponame_to_displayname(repo_name)

    if package_name is None:
        if repo_name is None:
            package_name = None
        else:
            package_name = str_to_identifier(repo_name)
    if install_requires is None:
        install_requires = ""
    else:
        install_requires = "\n    ".join(
            [
                x.strip()
                for x in install_requires.split(",")
                if x.strip() != "" and x.strip() not in ["pymemri", "pytest"]
            ]
        )

    if template_name == "classifier_plugin":
        assert package_name is not None
        assert user is not None
        repo_name_gitlab = str_to_gitlab_identifier(repo_name)

        # hacky, dont change!
        model_imports_ = (
            f"""
from transformers import pipeline, AutoTokenizer, AutoModelForSequenceClassification
from """
            + "pymemri.data.loader import load_huggingface_model_for_project"
        )
        model_init = f"""
        model = load_huggingface_model_for_project(project_path="{user}/{repo_name_gitlab}", client=client)
        tokenizer = AutoTokenizer.from_pretrained("distilroberta-base", model_max_length=512)
        self.pipeline = pipeline("sentiment-analysis", model=model, tokenizer=tokenizer, return_all_scores=True, truncation=True)
        """
        model_predict = """
        return self.pipeline(x)
        """
    else:
        model_imports_ = ""
        model_init = "raise NotImplementedError()"
        model_predict = "raise NotImplementedError()"

    return {
        "user": user,
        "package_name": package_name,
        "plugin_name": plugin_name,
        "repo_name": repo_name,
        "repo_url": repo_url,
        "description": str(description),
        "install_requires": install_requires,
        "model_imports": model_imports_,
        "model_init": model_init,
        "model_predict": model_predict,
    }


def _plugin_from_template(
    list_templates=False,
    user=None,
    repo_url=None,
    plugin_name=None,
    template_name="basic",
    package_name=None,
    description=None,
    target_dir=".",
    verbose=True,
    install_requires="",
    template_url=None,
):
    if list_templates:
        print("Available templates:")
        for template in get_templates():
            print(template)
        return

    template = download_plugin_template(
        template_name, **({"url": template_url} if template_url else {})
    )

    tgt_path = Path(target_dir)
    replace_dict = get_template_replace_dict(
        repo_url=repo_url,
        user=user,
        plugin_name=plugin_name,
        package_name=package_name,
        description=description,
        install_requires=install_requires,
        template_name=template_name,
    )
    logger.debug(replace_dict)

    formatter = TemplateFormatter(template, replace_dict, tgt_path)
    formatter.format()
    if verbose:
        formatter.print_filetree()

    logger.info(f"Created `{replace_dict['plugin_name']}` using the {template_name} template.")


@call_parse
def plugin_from_template(
    list_templates: Param("List available plugin templates", store_true) = False,
    user: Param("Your Gitlab username", str) = None,
    repo_url: Param("The url of your empty Gitlab plugin repository", str) = None,
    plugin_name: Param("Display name of your plugin", str) = None,
    template_name: Param(
        "Name of the template, use `list_templates` to see all available options"
    ) = "basic",
    package_name: Param("Name of your plugin python package", str) = None,
    description: Param("Description of your plugin", str) = None,
    target_dir: Param("Directory to output the formatted template", str) = ".",
    verbose: Param("Should print out dir", bool) = True,
    install_requires: Param(
        "Extra packages to install, provided as comma separated, e.g. pymemri,requests", str
    ) = "",
):
    """
    CLI that downloads and formats a plugin template according to the arguments, and local git repository.

    Args:
        list_templates (Param, optional): If True, only list available templates. Defaults to False.
        user (Param, optional): Your GitLab username. Defaults to None.
        repo_url (Param, optional): The url of your gitlab plugin repository. Defaults to None.
        plugin_name (Param, optional): The name of your plugin. Defaults to None.
        template_name (Param, optional): The name of the template used. To list all options, see `list_templates`.
            Defaults to "basic".
        package_name (Param, optional): The name of the python package of your plugin. Inferred if left blank. Defaults to None.
        description (Param, optional): An optional plugin description. Defaults to None.
        target_dir (Param, optional): Directory where the plugin template is generated. Defaults to ".".
    """
    _plugin_from_template(
        list_templates,
        user,
        repo_url,
        plugin_name,
        template_name,
        package_name,
        description,
        target_dir,
        verbose,
        install_requires,
    )



from pymemri.data.itembase import EdgeList, Item
from pymemri.data.schema import Person


class Dog(Item):
    properties = Item.properties + ["name", "age", "bites", "weight"]
    edges = Item.edges + ["friend"]

    def __init__(
        self,
        name: str = None,
        age: int = None,
        bites: bool = False,
        weight: float = None,
        friend: EdgeList[Person] = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.name = name
        self.age = age
        self.bites = bites
        self.weight = weight
        self.friend = EdgeList("friend", Person, friend)



from pathlib import Path
from typing import Any, List, Union

from . import _central_schema
from .itembase import EdgeList, Item


class OauthFlow(Item):
    description = """TBD"""
    properties = Item.properties + ["accessToken", "refreshToken", "accessTokenSecret"]
    edges = Item.edges + []

    def __init__(
        self,
        accessToken: str = None,
        refreshToken: str = None,
        accessTokenSecret: str = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.accessToken: Optional[str] = accessToken
        self.refreshToken: Optional[str] = refreshToken
        self.accessTokenSecret: Optional[str] = accessTokenSecret



import io
import math
from hashlib import sha256
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches
from matplotlib.collections import PatchCollection
from matplotlib.pyplot import imshow
from numpy.linalg import norm
from PIL import Image

from ._central_schema import File
from .basic import *
from .itembase import EdgeList, Item

DEFAULT_ENCODING = "PNG"


def show_images(images, cols=3, titles=None):
    image_list = [x.data for x in images] if isinstance(images[0], Photo) else images
    assert (titles is None) or (len(image_list) == len(titles))
    n_images = len(image_list)
    if titles is None:
        titles = ["" for i in range(1, n_images + 1)]
    fig = plt.figure()
    for n, (image, title) in enumerate(zip(image_list, titles)):
        a = fig.add_subplot(int(np.ceil(n_images / float(cols))), cols, n + 1)
        a.axis("off")
        if image.ndim == 2:
            plt.gray()
        plt.imshow(image[:, :, ::-1])
        a.set_title(title)
    fig.set_size_inches(np.array(fig.get_size_inches()) * n_images)
    plt.show()


class Photo(Item):

    properties = Item.properties + ["width", "height", "channels", "encoding", "mode"]
    edges = Item.edges + ["file"]

    def __init__(
        self,
        data: Any = None,
        includes: Any = None,  # TODO
        thumbnail: Any = None,  # TODO
        height: int = None,
        width: int = None,
        channels: int = None,
        encoding: str = None,
        mode: str = None,
        file: EdgeList[File] = None,
        _file_created: bool = False,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.private = ["data", "embedding", "path"]  # TODO
        self.height = height
        self.width = width
        self.channels = channels
        self.encoding = encoding
        self.mode = mode
        self.data = data
        self._file_created = _file_created

        self.file = EdgeList("file", "File", file)

    def show(self):
        fig, ax = plt.subplots(1)
        fig.set_figheight(15)
        fig.set_figwidth(15)
        ax.axis("off")
        imshow(self.data[:, :, ::-1])
        fig.set_size_inches((6, 6))
        plt.show()

    @property
    def size(self):
        return self.width, self.height

    @classmethod
    def from_data(cls, *args, **kwargs):
        res = super().from_data(*args, **kwargs)
        if res.file:
            res.file[0]  # TODO
        return res

    @classmethod
    def from_path(cls, path, size=None):
        pil_image = Image.open(path)
        encoding, mode, shape = cls.infer_PIL_metadata(pil_image)
        w, h, c = shape
        _bytes = cls.PIL_to_bytes(pil_image, encoding)

        res = cls(data=_bytes, height=h, width=w, channels=c, encoding=encoding, mode=mode)
        file = File.from_data(sha256=sha256(_bytes).hexdigest())
        res.add_edge("file", file)
        return res

    @classmethod
    def from_np(cls, data, size=None, *args, **kwargs):
        pil_image = Image.fromarray(data)
        if size is not None:
            pil_image = pil_image.resize(size)
        encoding, mode, shape = cls.infer_PIL_metadata(pil_image)
        w, h, c = shape
        _bytes = cls.PIL_to_bytes(pil_image, encoding)

        res = cls(data=_bytes, height=h, width=w, channels=c, encoding=encoding, mode=mode)
        file = File.from_data(sha256=sha256(_bytes).hexdigest())
        res.add_edge("file", file)
        return res

    @classmethod
    def from_bytes(cls, _bytes):
        image_stream = io.BytesIO(_bytes)
        pil_image = Image.open(image_stream)
        encoding, mode, shape = cls.infer_PIL_metadata(pil_image)
        w, h, c = shape

        res = cls(data=_bytes, height=h, width=w, channels=c, encoding=encoding, mode=mode)
        file = File.from_data(sha256=sha256(_bytes).hexdigest())
        res.add_edge("file", file)
        return res

    @staticmethod
    def PIL_to_bytes(pil_image, encoding):
        byte_io = io.BytesIO()
        pil_image.save(byte_io, encoding)
        return byte_io.getvalue()

    @staticmethod
    def infer_PIL_metadata(pil_image):
        encoding = pil_image.format or DEFAULT_ENCODING
        mode = pil_image.mode
        size = pil_image.size
        if len(size) == 3:
            w, h, c = size
        if len(size) == 2:
            w, h = size
            c = 1
        else:
            raise Error
        return encoding, mode, (w, h, c)

    def to_PIL(self):
        if self.data is None:
            raise ValueError("Photo object has no data")

        return Image.open(io.BytesIO(self.data))

    def to_np(self):
        pil_img = self.to_PIL()
        return np.asarray(pil_img)



# AUTOGENERATED, DO NOT EDIT!
# This file was generated by /tools/generate_central_schema.sh
# Visit https://gitlab.memri.io/memri/schema to learn more

from datetime import datetime
from typing import Any, Optional, Union

from pymemri.data.itembase import EdgeList, Item


class Account(Item):
    description = """An account or subscription, for instance for some online service, or a bank account or wallet."""
    properties = Item.properties + [
        "avatarUrl",
        "authEmail",
        "code",
        "displayName",
        "externalId",
        "handle",
        "identifier",
        "isMe",
        "itemType",
        "secret",
        "service",
        "accessToken",
        "refreshToken",
    ]
    edges = Item.edges + [
        "changelog",
        "cryptoTransaction",
        "location",
        "network",
        "ownCurrency",
        "owner",
        "trust",
        "profilePicture",
    ]

    def __init__(
        self,
        avatarUrl: str = None,
        authEmail: str = None,
        code: str = None,
        displayName: str = None,
        externalId: str = None,
        handle: str = None,
        identifier: str = None,
        isMe: bool = None,
        itemType: str = None,
        secret: str = None,
        service: str = None,
        accessToken: str = None,
        refreshToken: str = None,
        changelog: EdgeList["AuditItem"] = None,
        cryptoTransaction: EdgeList["CryptoTransaction"] = None,
        location: EdgeList["Location"] = None,
        network: EdgeList["Network"] = None,
        ownCurrency: EdgeList["CryptoCurrency"] = None,
        owner: EdgeList["Person"] = None,
        trust: EdgeList["Account"] = None,
        profilePicture: EdgeList["Photo"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.avatarUrl: Optional[str] = avatarUrl
        self.authEmail: Optional[str] = authEmail
        self.code: Optional[str] = code
        self.displayName: Optional[str] = displayName
        self.externalId: Optional[str] = externalId
        self.handle: Optional[str] = handle
        self.identifier: Optional[str] = identifier
        self.isMe: Optional[bool] = isMe
        self.itemType: Optional[str] = itemType
        self.secret: Optional[str] = secret
        self.service: Optional[str] = service
        self.accessToken: Optional[str] = accessToken
        self.refreshToken: Optional[str] = refreshToken

        # Edges
        self.changelog: EdgeList["AuditItem"] = EdgeList("changelog", "AuditItem", changelog)
        self.cryptoTransaction: EdgeList["CryptoTransaction"] = EdgeList(
            "cryptoTransaction", "CryptoTransaction", cryptoTransaction
        )
        self.location: EdgeList["Location"] = EdgeList("location", "Location", location)
        self.network: EdgeList["Network"] = EdgeList("network", "Network", network)
        self.ownCurrency: EdgeList["CryptoCurrency"] = EdgeList(
            "ownCurrency", "CryptoCurrency", ownCurrency
        )
        self.owner: EdgeList["Person"] = EdgeList("owner", "Person", owner)
        self.trust: EdgeList["Account"] = EdgeList("trust", "Account", trust)
        self.profilePicture: EdgeList["Photo"] = EdgeList("profilePicture", "Photo", profilePicture)


class AuditItem(Item):
    description = """TBD"""
    properties = Item.properties + ["actionname", "content", "date"]
    edges = Item.edges + []

    def __init__(
        self, actionname: str = None, content: str = None, date: datetime = None, **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.actionname: Optional[str] = actionname
        self.content: Optional[str] = content
        self.date: Optional[datetime] = date


class CVUStoredDefinition(Item):
    description = """TBD"""
    properties = Item.properties + [
        "definition",
        "domain",
        "itemType",
        "name",
        "queryStr",
        "renderer",
        "selector",
        "definitionType",
    ]
    edges = Item.edges + []

    def __init__(
        self,
        definition: str = None,
        domain: str = None,
        itemType: str = None,
        name: str = None,
        queryStr: str = None,
        renderer: str = None,
        selector: str = None,
        definitionType: str = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.definition: Optional[str] = definition
        self.domain: Optional[str] = domain
        self.itemType: Optional[str] = itemType
        self.name: Optional[str] = name
        self.queryStr: Optional[str] = queryStr
        self.renderer: Optional[str] = renderer
        self.selector: Optional[str] = selector
        self.definitionType: Optional[str] = definitionType


class CategoricalPrediction(Item):
    description = """"""
    properties = Item.properties + ["source", "value", "probs"]
    edges = Item.edges + ["model"]

    def __init__(
        self,
        source: str = None,
        value: str = None,
        probs: str = None,
        model: EdgeList["Model"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.source: Optional[str] = source
        self.value: Optional[str] = value
        self.probs: Optional[str] = probs

        # Edges
        self.model: EdgeList["Model"] = EdgeList("model", "Model", model)


class CreativeWork(Item):
    description = """The most generic kind of creative work, including books, movies, photographs, software programs, etc."""
    properties = Item.properties + [
        "abstract",
        "content",
        "datePublished",
        "itemType",
        "keyword",
        "textContent",
        "title",
        "transcript",
    ]
    edges = Item.edges + ["contentLocation", "file", "locationCreated", "writtenBy"]

    def __init__(
        self,
        abstract: str = None,
        content: str = None,
        datePublished: datetime = None,
        itemType: str = None,
        keyword: str = None,
        textContent: str = None,
        title: str = None,
        transcript: str = None,
        contentLocation: EdgeList["Location"] = None,
        file: EdgeList["File"] = None,
        locationCreated: EdgeList["Location"] = None,
        writtenBy: EdgeList["Person"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.abstract: Optional[str] = abstract
        self.content: Optional[str] = content
        self.datePublished: Optional[datetime] = datePublished
        self.itemType: Optional[str] = itemType
        self.keyword: Optional[str] = keyword
        self.textContent: Optional[str] = textContent
        self.title: Optional[str] = title
        self.transcript: Optional[str] = transcript

        # Edges
        self.contentLocation: EdgeList["Location"] = EdgeList(
            "contentLocation", "Location", contentLocation
        )
        self.file: EdgeList["File"] = EdgeList("file", "File", file)
        self.locationCreated: EdgeList["Location"] = EdgeList(
            "locationCreated", "Location", locationCreated
        )
        self.writtenBy: EdgeList["Person"] = EdgeList("writtenBy", "Person", writtenBy)


class CryptoCurrency(Item):
    description = """"""
    properties = Item.properties + ["myToken", "name", "topic"]
    edges = Item.edges + ["currencySetting", "picture"]

    def __init__(
        self,
        myToken: float = None,
        name: str = None,
        topic: str = None,
        currencySetting: EdgeList["CurrencySetting"] = None,
        picture: EdgeList["Photo"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.myToken: Optional[float] = myToken
        self.name: Optional[str] = name
        self.topic: Optional[str] = topic

        # Edges
        self.currencySetting: EdgeList["CurrencySetting"] = EdgeList(
            "currencySetting", "CurrencySetting", currencySetting
        )
        self.picture: EdgeList["Photo"] = EdgeList("picture", "Photo", picture)


class CryptoKey(Item):
    description = """A key used in an cryptography protocol."""
    properties = Item.properties + [
        "active",
        "itemType",
        "keystr",
        "name",
        "role",
        "starred",
    ]
    edges = Item.edges + ["owner"]

    def __init__(
        self,
        active: bool = None,
        itemType: str = None,
        keystr: str = None,
        name: str = None,
        role: str = None,
        starred: bool = None,
        owner: EdgeList["Person"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.active: Optional[bool] = active
        self.itemType: Optional[str] = itemType
        self.keystr: Optional[str] = keystr
        self.name: Optional[str] = name
        self.role: Optional[str] = role
        self.starred: Optional[bool] = starred

        # Edges
        self.owner: EdgeList["Person"] = EdgeList("owner", "Person", owner)


class CryptoTransaction(Item):
    description = """"""
    properties = Item.properties + ["outward", "quantity"]
    edges = Item.edges + ["cryptoCurrency", "relateToOther", "relateToOwner"]

    def __init__(
        self,
        outward: bool = None,
        quantity: float = None,
        cryptoCurrency: EdgeList["CryptoCurrency"] = None,
        relateToOther: EdgeList["Account"] = None,
        relateToOwner: EdgeList["Account"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.outward: Optional[bool] = outward
        self.quantity: Optional[float] = quantity

        # Edges
        self.cryptoCurrency: EdgeList["CryptoCurrency"] = EdgeList(
            "cryptoCurrency", "CryptoCurrency", cryptoCurrency
        )
        self.relateToOther: EdgeList["Account"] = EdgeList(
            "relateToOther", "Account", relateToOther
        )
        self.relateToOwner: EdgeList["Account"] = EdgeList(
            "relateToOwner", "Account", relateToOwner
        )


class CurrencySetting(Item):
    description = """"""
    properties = Item.properties + [
        "deviceAddress",
        "profileAddress",
        "seedPhrase",
        "tokenAddress",
    ]
    edges = Item.edges + ["wallet"]

    def __init__(
        self,
        deviceAddress: str = None,
        profileAddress: str = None,
        seedPhrase: str = None,
        tokenAddress: str = None,
        wallet: EdgeList["Wallet"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.deviceAddress: Optional[str] = deviceAddress
        self.profileAddress: Optional[str] = profileAddress
        self.seedPhrase: Optional[str] = seedPhrase
        self.tokenAddress: Optional[str] = tokenAddress

        # Edges
        self.wallet: EdgeList["Wallet"] = EdgeList("wallet", "Wallet", wallet)


class Dataset(Item):
    description = """A Dataset of items."""
    properties = Item.properties + ["name", "queryStr"]
    edges = Item.edges + ["entry", "feature", "labellingTask", "datasetType"]

    def __init__(
        self,
        name: str = None,
        queryStr: str = None,
        entry: EdgeList["DatasetEntry"] = None,
        feature: EdgeList["ItemPropertySchema"] = None,
        labellingTask: EdgeList["LabellingTask"] = None,
        datasetType: EdgeList["DatasetType"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.name: Optional[str] = name
        self.queryStr: Optional[str] = queryStr

        # Edges
        self.entry: EdgeList["DatasetEntry"] = EdgeList("entry", "DatasetEntry", entry)
        self.feature: EdgeList["ItemPropertySchema"] = EdgeList(
            "feature", "ItemPropertySchema", feature
        )
        self.labellingTask: EdgeList["LabellingTask"] = EdgeList(
            "labellingTask", "LabellingTask", labellingTask
        )
        self.datasetType: EdgeList["DatasetType"] = EdgeList(
            "datasetType", "DatasetType", datasetType
        )


class DatasetEntry(Item):
    description = """Entry item of dataset."""
    properties = Item.properties + ["skippedByLabeller"]
    edges = Item.edges + ["data", "annotation"]

    def __init__(
        self,
        skippedByLabeller: bool = None,
        data: EdgeList[Union["Message", "Tweet", "EmailMessage"]] = None,
        annotation: EdgeList["CategoricalLabel"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.skippedByLabeller: Optional[bool] = skippedByLabeller

        # Edges
        self.data: EdgeList[Union["Message", "Tweet", "EmailMessage"]] = EdgeList(
            "data", Union["Message", "Tweet", "EmailMessage"], data
        )
        self.annotation: EdgeList["CategoricalLabel"] = EdgeList(
            "annotation", "CategoricalLabel", annotation
        )


class DatasetType(Item):
    description = """Fixed dictionary for datasets."""
    properties = Item.properties + ["name", "queryStr"]
    edges = Item.edges + ["datasetPlugin"]

    def __init__(
        self,
        name: str = None,
        queryStr: str = None,
        datasetPlugin: EdgeList["Plugin"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.name: Optional[str] = name
        self.queryStr: Optional[str] = queryStr

        # Edges
        self.datasetPlugin: EdgeList["Plugin"] = EdgeList("datasetPlugin", "Plugin", datasetPlugin)


class Diet(Item):
    description = """A strategy of regulating the intake of food to achieve or maintain a specific health-related goal."""
    properties = Item.properties + [
        "abstract",
        "content",
        "datePublished",
        "duration",
        "itemType",
        "keyword",
        "name",
        "textContent",
        "title",
        "transcript",
    ]
    edges = Item.edges + ["contentLocation", "file", "locationCreated", "writtenBy"]

    def __init__(
        self,
        abstract: str = None,
        content: str = None,
        datePublished: datetime = None,
        duration: int = None,
        itemType: str = None,
        keyword: str = None,
        name: str = None,
        textContent: str = None,
        title: str = None,
        transcript: str = None,
        contentLocation: EdgeList["Location"] = None,
        file: EdgeList["File"] = None,
        locationCreated: EdgeList["Location"] = None,
        writtenBy: EdgeList["Person"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.abstract: Optional[str] = abstract
        self.content: Optional[str] = content
        self.datePublished: Optional[datetime] = datePublished
        self.duration: Optional[int] = duration
        self.itemType: Optional[str] = itemType
        self.keyword: Optional[str] = keyword
        self.name: Optional[str] = name
        self.textContent: Optional[str] = textContent
        self.title: Optional[str] = title
        self.transcript: Optional[str] = transcript

        # Edges
        self.contentLocation: EdgeList["Location"] = EdgeList(
            "contentLocation", "Location", contentLocation
        )
        self.file: EdgeList["File"] = EdgeList("file", "File", file)
        self.locationCreated: EdgeList["Location"] = EdgeList(
            "locationCreated", "Location", locationCreated
        )
        self.writtenBy: EdgeList["Person"] = EdgeList("writtenBy", "Person", writtenBy)


class File(Item):
    description = """Any file that can be stored on disk."""
    properties = Item.properties + ["filename", "keystr", "nonce", "sha256", "starred"]
    edges = Item.edges + []

    def __init__(
        self,
        filename: str = None,
        keystr: str = None,
        nonce: str = None,
        sha256: str = None,
        starred: bool = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.filename: Optional[str] = filename
        self.keystr: Optional[str] = keystr
        self.nonce: Optional[str] = nonce
        self.sha256: Optional[str] = sha256
        self.starred: Optional[bool] = starred


class Integrator(Item):
    description = """An integrator operates on your database enhances your personal data by inferring facts over existing data and adding those to the database."""
    properties = Item.properties + ["name", "repository"]
    edges = Item.edges + []

    def __init__(self, name: str = None, repository: str = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.name: Optional[str] = name
        self.repository: Optional[str] = repository


class ItemEdgeSchema(Item):
    description = """"""
    properties = Item.properties + ["edgeName", "sourceType", "targetType"]
    edges = Item.edges + []

    def __init__(
        self, edgeName: str = None, sourceType: str = None, targetType: str = None, **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.edgeName: Optional[str] = edgeName
        self.sourceType: Optional[str] = sourceType
        self.targetType: Optional[str] = targetType


class ItemPropertySchema(Item):
    description = """"""
    properties = Item.properties + ["itemType", "propertyName", "valueType"]
    edges = Item.edges + []

    def __init__(
        self, itemType: str = None, propertyName: str = None, valueType: str = None, **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.itemType: Optional[str] = itemType
        self.propertyName: Optional[str] = propertyName
        self.valueType: Optional[str] = valueType


class Label(Item):
    description = """Attached to an Item, to mark it to be something."""
    properties = Item.properties + ["color", "name"]
    edges = Item.edges + []

    def __init__(self, color: str = None, name: str = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.color: Optional[str] = color
        self.name: Optional[str] = name


class LabelAnnotation(Item):
    description = """"""
    properties = Item.properties + ["allowSharing", "isSubmitted"]
    edges = Item.edges + []

    def __init__(self, allowSharing: bool = None, isSubmitted: bool = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.allowSharing: Optional[bool] = allowSharing
        self.isSubmitted: Optional[bool] = isSubmitted


class LabelOption(Item):
    description = """Attached to an Item, to mark it to be something."""
    properties = Item.properties + ["color", "name"]
    edges = Item.edges + []

    def __init__(self, color: str = None, name: str = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.color: Optional[str] = color
        self.name: Optional[str] = name


class LabellingDataType(Item):
    description = """A labelling data type definition."""
    properties = Item.properties + ["name", "labelType"]
    edges = Item.edges + []

    def __init__(self, name: str = None, labelType: str = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.name: Optional[str] = name
        self.labelType: Optional[str] = labelType


class LabellingTask(Item):
    description = """A labelling task definition."""
    properties = Item.properties + ["name", "currentLabelOption"]
    edges = Item.edges + ["taskType", "labelOption", "view"]

    def __init__(
        self,
        name: str = None,
        currentLabelOption: str = None,
        taskType: EdgeList["TextClassification"] = None,
        labelOption: EdgeList["LabelOption"] = None,
        view: EdgeList["CVUStoredDefinition"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.name: Optional[str] = name
        self.currentLabelOption: Optional[str] = currentLabelOption

        # Edges
        self.taskType: EdgeList["TextClassification"] = EdgeList(
            "taskType", "TextClassification", taskType
        )
        self.labelOption: EdgeList["LabelOption"] = EdgeList(
            "labelOption", "LabelOption", labelOption
        )
        self.view: EdgeList["CVUStoredDefinition"] = EdgeList("view", "CVUStoredDefinition", view)


class Location(Item):
    description = """The location of something."""
    properties = Item.properties + ["latitude", "longitude"]
    edges = Item.edges + []

    def __init__(self, latitude: float = None, longitude: float = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.latitude: Optional[float] = latitude
        self.longitude: Optional[float] = longitude


class MediaObject(Item):
    description = """A media object, such as an image, video, or audio object embedded in a web page or a downloadable dataset i.e. DataDownload. Note that a creative work may have many media objects associated with it on the same web page. For example, a page about a single song (MusicRecording) may have a music video (VideoObject), and a high and low bandwidth audio stream (2 AudioObject's)."""
    properties = Item.properties + [
        "bitrate",
        "duration",
        "endTime",
        "fileLocation",
        "startTime",
    ]
    edges = Item.edges + ["file", "includes"]

    def __init__(
        self,
        bitrate: int = None,
        duration: int = None,
        endTime: datetime = None,
        fileLocation: str = None,
        startTime: datetime = None,
        file: EdgeList["File"] = None,
        includes: EdgeList["Person"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.bitrate: Optional[int] = bitrate
        self.duration: Optional[int] = duration
        self.endTime: Optional[datetime] = endTime
        self.fileLocation: Optional[str] = fileLocation
        self.startTime: Optional[datetime] = startTime

        # Edges
        self.file: EdgeList["File"] = EdgeList("file", "File", file)
        self.includes: EdgeList["Person"] = EdgeList("includes", "Person", includes)


class MedicalCondition(Item):
    description = """Any condition of the human body that affects the normal functioning of a person, whether physically or mentally. Includes diseases, injuries, disabilities, disorders, syndromes, etc."""
    properties = Item.properties + ["conditiontype", "itemType", "name"]
    edges = Item.edges + []

    def __init__(self, conditiontype: str = None, itemType: str = None, name: str = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.conditiontype: Optional[str] = conditiontype
        self.itemType: Optional[str] = itemType
        self.name: Optional[str] = name


class MessageChannel(Item):
    description = """A chat is a collection of messages."""
    properties = Item.properties + ["encrypted", "externalId", "name", "topic"]
    edges = Item.edges + ["photo", "receiver"]

    def __init__(
        self,
        encrypted: bool = None,
        externalId: str = None,
        name: str = None,
        topic: str = None,
        photo: EdgeList["Photo"] = None,
        receiver: EdgeList["Account"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.encrypted: Optional[bool] = encrypted
        self.externalId: Optional[str] = externalId
        self.name: Optional[str] = name
        self.topic: Optional[str] = topic

        # Edges
        self.photo: EdgeList["Photo"] = EdgeList("photo", "Photo", photo)
        self.receiver: EdgeList["Account"] = EdgeList("receiver", "Account", receiver)


class Model(Item):
    description = """A Machine learning model"""
    properties = Item.properties + ["name", "version"]
    edges = Item.edges + []

    def __init__(self, name: str = None, version: str = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.name: Optional[str] = name
        self.version: Optional[str] = version


class NavigationItem(Item):
    description = """TBD"""
    properties = Item.properties + [
        "icon",
        "itemType",
        "sequence",
        "sessionName",
        "title",
    ]
    edges = Item.edges + []

    def __init__(
        self,
        icon: str = None,
        itemType: str = None,
        sequence: int = None,
        sessionName: str = None,
        title: str = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.icon: Optional[str] = icon
        self.itemType: Optional[str] = itemType
        self.sequence: Optional[int] = sequence
        self.sessionName: Optional[str] = sessionName
        self.title: Optional[str] = title


class Network(Item):
    description = (
        """A group or system of interconnected people or things, for instance a social network."""
    )
    properties = Item.properties + ["name"]
    edges = Item.edges + ["website"]

    def __init__(self, name: str = None, website: EdgeList["Website"] = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.name: Optional[str] = name

        # Edges
        self.website: EdgeList["Website"] = EdgeList("website", "Website", website)


class OauthFlow(Item):
    description = """"""
    properties = Item.properties + [
        "accessToken",
        "accessTokenSecret",
        "refreshToken",
        "service",
    ]
    edges = Item.edges + []

    def __init__(
        self,
        accessToken: str = None,
        accessTokenSecret: str = None,
        refreshToken: str = None,
        service: str = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.accessToken: Optional[str] = accessToken
        self.accessTokenSecret: Optional[str] = accessTokenSecret
        self.refreshToken: Optional[str] = refreshToken
        self.service: Optional[str] = service


class Person(Item):
    description = """A person (alive, dead, undead, or fictional)."""
    properties = Item.properties + [
        "addressBookId",
        "birthDate",
        "deathDate",
        "displayName",
        "email",
        "firstName",
        "gender",
        "lastName",
        "role",
        "sexualOrientation",
        "starred",
    ]
    edges = Item.edges + [
        "account",
        "address",
        "birthPlace",
        "cryptoKey",
        "deathPlace",
        "diet",
        "hasPhoneNumber",
        "label",
        "me",
        "medicalCondition",
        "mergedFrom",
        "profilePicture",
        "relationship",
        "website",
    ]

    def __init__(
        self,
        addressBookId: str = None,
        birthDate: datetime = None,
        deathDate: datetime = None,
        displayName: str = None,
        email: str = None,
        firstName: str = None,
        gender: str = None,
        lastName: str = None,
        role: str = None,
        sexualOrientation: str = None,
        starred: bool = None,
        account: EdgeList["Account"] = None,
        address: EdgeList["Address"] = None,
        birthPlace: EdgeList["Location"] = None,
        cryptoKey: EdgeList["CryptoKey"] = None,
        deathPlace: EdgeList["Location"] = None,
        diet: EdgeList["Diet"] = None,
        hasPhoneNumber: EdgeList["PhoneNumber"] = None,
        label: EdgeList["Label"] = None,
        me: EdgeList["Person"] = None,
        medicalCondition: EdgeList["MedicalCondition"] = None,
        mergedFrom: EdgeList["Person"] = None,
        profilePicture: EdgeList["Photo"] = None,
        relationship: EdgeList["Relationship"] = None,
        website: EdgeList["Website"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.addressBookId: Optional[str] = addressBookId
        self.birthDate: Optional[datetime] = birthDate
        self.deathDate: Optional[datetime] = deathDate
        self.displayName: Optional[str] = displayName
        self.email: Optional[str] = email
        self.firstName: Optional[str] = firstName
        self.gender: Optional[str] = gender
        self.lastName: Optional[str] = lastName
        self.role: Optional[str] = role
        self.sexualOrientation: Optional[str] = sexualOrientation
        self.starred: Optional[bool] = starred

        # Edges
        self.account: EdgeList["Account"] = EdgeList("account", "Account", account)
        self.address: EdgeList["Address"] = EdgeList("address", "Address", address)
        self.birthPlace: EdgeList["Location"] = EdgeList("birthPlace", "Location", birthPlace)
        self.cryptoKey: EdgeList["CryptoKey"] = EdgeList("cryptoKey", "CryptoKey", cryptoKey)
        self.deathPlace: EdgeList["Location"] = EdgeList("deathPlace", "Location", deathPlace)
        self.diet: EdgeList["Diet"] = EdgeList("diet", "Diet", diet)
        self.hasPhoneNumber: EdgeList["PhoneNumber"] = EdgeList(
            "hasPhoneNumber", "PhoneNumber", hasPhoneNumber
        )
        self.label: EdgeList["Label"] = EdgeList("label", "Label", label)
        self.me: EdgeList["Person"] = EdgeList("me", "Person", me)
        self.medicalCondition: EdgeList["MedicalCondition"] = EdgeList(
            "medicalCondition", "MedicalCondition", medicalCondition
        )
        self.mergedFrom: EdgeList["Person"] = EdgeList("mergedFrom", "Person", mergedFrom)
        self.profilePicture: EdgeList["Photo"] = EdgeList("profilePicture", "Photo", profilePicture)
        self.relationship: EdgeList["Relationship"] = EdgeList(
            "relationship", "Relationship", relationship
        )
        self.website: EdgeList["Website"] = EdgeList("website", "Website", website)


class PhoneNumber(Item):
    description = """A telephone number, SIP Address."""
    properties = Item.properties + ["phoneNumber"]
    edges = Item.edges + []

    def __init__(self, phoneNumber: str = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.phoneNumber: Optional[str] = phoneNumber


class Plugin(Item):
    description = """Information about a Plugin"""
    properties = Item.properties + [
        "bundleImage",
        "containerImage",
        "configJson",
        "config",
        "dataType",
        "icon",
        "pluginDescription",
        "name",
        "pluginModule",
        "pluginName",
        "pluginType",
        "gitProjectId",
    ]
    edges = Item.edges + ["view"]

    def __init__(
        self,
        bundleImage: str = None,
        containerImage: str = None,
        configJson: str = None,
        config: str = None,
        dataType: str = None,
        icon: str = None,
        pluginDescription: str = None,
        name: str = None,
        pluginModule: str = None,
        pluginName: str = None,
        pluginType: str = None,
        gitProjectId: int = None,
        view: EdgeList["CVUStoredDefinition"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.bundleImage: Optional[str] = bundleImage
        self.containerImage: Optional[str] = containerImage
        self.configJson: Optional[str] = configJson
        self.config: Optional[str] = config
        self.dataType: Optional[str] = dataType
        self.icon: Optional[str] = icon
        self.pluginDescription: Optional[str] = pluginDescription
        self.name: Optional[str] = name
        self.pluginModule: Optional[str] = pluginModule
        self.pluginName: Optional[str] = pluginName
        self.pluginType: Optional[str] = pluginType
        self.gitProjectId: Optional[int] = gitProjectId

        # Edges
        self.view: EdgeList["CVUStoredDefinition"] = EdgeList("view", "CVUStoredDefinition", view)


class Post(Item):
    description = """Post from social media"""
    properties = Item.properties + [
        "externalId",
        "message",
        "postDate",
        "postType",
        "isMock",
    ]
    edges = Item.edges + ["author", "comment", "parent", "photo"]

    def __init__(
        self,
        externalId: str = None,
        message: str = None,
        postDate: datetime = None,
        postType: str = None,
        isMock: bool = None,
        author: EdgeList["Account"] = None,
        comment: EdgeList["Post"] = None,
        parent: EdgeList["Post"] = None,
        photo: EdgeList["Photo"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.externalId: Optional[str] = externalId
        self.message: Optional[str] = message
        self.postDate: Optional[datetime] = postDate
        self.postType: Optional[str] = postType
        self.isMock: Optional[bool] = isMock

        # Edges
        self.author: EdgeList["Account"] = EdgeList("author", "Account", author)
        self.comment: EdgeList["Post"] = EdgeList("comment", "Post", comment)
        self.parent: EdgeList["Post"] = EdgeList("parent", "Post", parent)
        self.photo: EdgeList["Photo"] = EdgeList("photo", "Photo", photo)


class Project(Item):
    description = """Labelling project"""
    properties = Item.properties + ["name", "gitlabUrl", "dataSource"]
    edges = Item.edges + ["dataset", "labellingPlugin"]

    def __init__(
        self,
        name: str = None,
        gitlabUrl: str = None,
        dataSource: str = None,
        dataset: EdgeList["Dataset"] = None,
        labellingPlugin: EdgeList["Plugin"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.name: Optional[str] = name
        self.gitlabUrl: Optional[str] = gitlabUrl
        self.dataSource: Optional[str] = dataSource

        # Edges
        self.dataset: EdgeList["Dataset"] = EdgeList("dataset", "Dataset", dataset)
        self.labellingPlugin: EdgeList["Plugin"] = EdgeList(
            "labellingPlugin", "Plugin", labellingPlugin
        )


class Receipt(Item):
    description = """A bill that describes money owed for some Transaction."""
    properties = Item.properties + ["category", "store", "totalCost"]
    edges = Item.edges + ["file", "photo"]

    def __init__(
        self,
        category: str = None,
        store: str = None,
        totalCost: float = None,
        file: EdgeList["File"] = None,
        photo: EdgeList["Photo"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.category: Optional[str] = category
        self.store: Optional[str] = store
        self.totalCost: Optional[float] = totalCost

        # Edges
        self.file: EdgeList["File"] = EdgeList("file", "File", file)
        self.photo: EdgeList["Photo"] = EdgeList("photo", "Photo", photo)


class Relationship(Item):
    description = """Relation of people, that indicates type of relationship and its value"""
    properties = Item.properties + ["label", "proximityValue"]
    edges = Item.edges + ["relationship"]

    def __init__(
        self,
        label: str = None,
        proximityValue: int = None,
        relationship: EdgeList["Person"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.label: Optional[str] = label
        self.proximityValue: Optional[int] = proximityValue

        # Edges
        self.relationship: EdgeList["Person"] = EdgeList("relationship", "Person", relationship)


class Setting(Item):
    description = """A setting, named by a key, specifications in JSON format."""
    properties = Item.properties + ["json", "keystr"]
    edges = Item.edges + []

    def __init__(self, json: str = None, keystr: str = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.json: Optional[str] = json
        self.keystr: Optional[str] = keystr


class SuggestedMerge(Item):
    description = """Describes a suggestion to merge two or more items"""
    properties = Item.properties + ["score", "task"]
    edges = Item.edges + ["mergeFrom"]

    def __init__(
        self, score: float = None, task: str = None, mergeFrom: EdgeList["Person"] = None, **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.score: Optional[float] = score
        self.task: Optional[str] = task

        # Edges
        self.mergeFrom: EdgeList["Person"] = EdgeList("mergeFrom", "Person", mergeFrom)


class VoteAction(Item):
    description = """The act casting a vote."""
    properties = Item.properties + ["dateExecuted"]
    edges = Item.edges + []

    def __init__(self, dateExecuted: datetime = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.dateExecuted: Optional[datetime] = dateExecuted


class Wallet(Item):
    description = """"""
    properties = Item.properties + ["name"]
    edges = Item.edges + ["picture"]

    def __init__(self, name: str = None, picture: EdgeList["Photo"] = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.name: Optional[str] = name

        # Edges
        self.picture: EdgeList["Photo"] = EdgeList("picture", "Photo", picture)


class Website(Item):
    description = """A Website is a set of related web pages and other items typically served from a single web domain and accessible via URLs."""
    properties = Item.properties + ["itemType", "url"]
    edges = Item.edges + []

    def __init__(self, itemType: str = None, url: str = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.itemType: Optional[str] = itemType
        self.url: Optional[str] = url


class WrittenWork(CreativeWork):
    description = (
        """A written work, for instance a book, article or note. Doesn't have to be published."""
    )
    properties = CreativeWork.properties + []
    edges = CreativeWork.edges + []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class Indexer(Integrator):
    description = """An indexer enhances your personal data by inferring facts over existing data and adding those to the database."""
    properties = Integrator.properties + [
        "bundleImage",
        "icon",
        "indexerClass",
        "itemDescription",
        "queryStr",
        "runDestination",
    ]
    edges = Integrator.edges + ["indexerRun"]

    def __init__(
        self,
        bundleImage: str = None,
        icon: str = None,
        indexerClass: str = None,
        itemDescription: str = None,
        queryStr: str = None,
        runDestination: str = None,
        indexerRun: EdgeList["IndexerRun"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.bundleImage: Optional[str] = bundleImage
        self.icon: Optional[str] = icon
        self.indexerClass: Optional[str] = indexerClass
        self.itemDescription: Optional[str] = itemDescription
        self.queryStr: Optional[str] = queryStr
        self.runDestination: Optional[str] = runDestination

        # Edges
        self.indexerRun: EdgeList["IndexerRun"] = EdgeList("indexerRun", "IndexerRun", indexerRun)


class IndexerRun(Integrator):
    description = """A run of a certain Indexer."""
    properties = Integrator.properties + [
        "errorMessage",
        "progress",
        "progressMessage",
        "queryStr",
        "runStatus",
        "targetDataType",
    ]
    edges = Integrator.edges + ["indexer"]

    def __init__(
        self,
        errorMessage: str = None,
        progress: float = None,
        progressMessage: str = None,
        queryStr: str = None,
        runStatus: str = None,
        targetDataType: str = None,
        indexer: EdgeList["Indexer"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.errorMessage: Optional[str] = errorMessage
        self.progress: Optional[float] = progress
        self.progressMessage: Optional[str] = progressMessage
        self.queryStr: Optional[str] = queryStr
        self.runStatus: Optional[str] = runStatus
        self.targetDataType: Optional[str] = targetDataType

        # Edges
        self.indexer: EdgeList["Indexer"] = EdgeList("indexer", "Indexer", indexer)


class CategoricalLabel(LabelAnnotation):
    description = """"""
    properties = LabelAnnotation.properties + ["labelValue"]
    edges = LabelAnnotation.edges + []

    def __init__(self, labelValue: str = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.labelValue: Optional[str] = labelValue


class TextClassification(LabellingDataType):
    description = """Text classification option"""
    properties = LabellingDataType.properties + []
    edges = LabellingDataType.edges + []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class Address(Location):
    description = """A postal address."""
    properties = Location.properties + [
        "city",
        "itemType",
        "locationAutoLookupHash",
        "postalCode",
        "state",
        "street",
    ]
    edges = Location.edges + ["changelog", "country", "location"]

    def __init__(
        self,
        city: str = None,
        itemType: str = None,
        locationAutoLookupHash: str = None,
        postalCode: str = None,
        state: str = None,
        street: str = None,
        changelog: EdgeList["AuditItem"] = None,
        country: EdgeList["Country"] = None,
        location: EdgeList["Location"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.city: Optional[str] = city
        self.itemType: Optional[str] = itemType
        self.locationAutoLookupHash: Optional[str] = locationAutoLookupHash
        self.postalCode: Optional[str] = postalCode
        self.state: Optional[str] = state
        self.street: Optional[str] = street

        # Edges
        self.changelog: EdgeList["AuditItem"] = EdgeList("changelog", "AuditItem", changelog)
        self.country: EdgeList["Country"] = EdgeList("country", "Country", country)
        self.location: EdgeList["Location"] = EdgeList("location", "Location", location)


class Country(Location):
    description = """A country."""
    properties = Location.properties + ["name"]
    edges = Location.edges + ["flag", "location"]

    def __init__(
        self,
        name: str = None,
        flag: EdgeList["Photo"] = None,
        location: EdgeList["Location"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.name: Optional[str] = name

        # Edges
        self.flag: EdgeList["Photo"] = EdgeList("flag", "Photo", flag)
        self.location: EdgeList["Location"] = EdgeList("location", "Location", location)


class Photo(MediaObject):
    description = """An image file."""
    properties = MediaObject.properties + ["caption", "exifData", "name"]
    edges = MediaObject.edges + ["changelog", "label", "thumbnail"]

    def __init__(
        self,
        caption: str = None,
        exifData: str = None,
        name: str = None,
        changelog: EdgeList["AuditItem"] = None,
        label: EdgeList["Label"] = None,
        thumbnail: EdgeList["File"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.caption: Optional[str] = caption
        self.exifData: Optional[str] = exifData
        self.name: Optional[str] = name

        # Edges
        self.changelog: EdgeList["AuditItem"] = EdgeList("changelog", "AuditItem", changelog)
        self.label: EdgeList["Label"] = EdgeList("label", "Label", label)
        self.thumbnail: EdgeList["File"] = EdgeList("thumbnail", "File", thumbnail)


class Tweet(Post):
    description = """A Twitter Tweet"""
    properties = Post.properties + ["service"]
    edges = Post.edges + ["mention"]

    def __init__(self, service: str = None, mention: EdgeList["Account"] = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.service: Optional[str] = service

        # Edges
        self.mention: EdgeList["Account"] = EdgeList("mention", "Account", mention)


class Message(WrittenWork):
    description = """A single message."""
    properties = WrittenWork.properties + [
        "dateReceived",
        "dateSent",
        "externalId",
        "service",
        "subject",
        "sourceProject",
        "isMock",
    ]
    edges = WrittenWork.edges + [
        "message",
        "messageChannel",
        "photo",
        "receiver",
        "sender",
        "label",
    ]

    def __init__(
        self,
        dateReceived: datetime = None,
        dateSent: datetime = None,
        externalId: str = None,
        service: str = None,
        subject: str = None,
        sourceProject: str = None,
        isMock: bool = None,
        message: EdgeList["Message"] = None,
        messageChannel: EdgeList["MessageChannel"] = None,
        photo: EdgeList["Photo"] = None,
        receiver: EdgeList["Account"] = None,
        sender: EdgeList["Account"] = None,
        label: EdgeList["CategoricalPrediction"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.dateReceived: Optional[datetime] = dateReceived
        self.dateSent: Optional[datetime] = dateSent
        self.externalId: Optional[str] = externalId
        self.service: Optional[str] = service
        self.subject: Optional[str] = subject
        self.sourceProject: Optional[str] = sourceProject
        self.isMock: Optional[bool] = isMock

        # Edges
        self.message: EdgeList["Message"] = EdgeList("message", "Message", message)
        self.messageChannel: EdgeList["MessageChannel"] = EdgeList(
            "messageChannel", "MessageChannel", messageChannel
        )
        self.photo: EdgeList["Photo"] = EdgeList("photo", "Photo", photo)
        self.receiver: EdgeList["Account"] = EdgeList("receiver", "Account", receiver)
        self.sender: EdgeList["Account"] = EdgeList("sender", "Account", sender)
        self.label: EdgeList["CategoricalPrediction"] = EdgeList(
            "label", "CategoricalPrediction", label
        )


class Note(WrittenWork):
    description = """A file containing a note."""
    properties = WrittenWork.properties + ["starred"]
    edges = WrittenWork.edges + ["label"]

    def __init__(self, starred: bool = None, label: EdgeList["Label"] = None, **kwargs):
        super().__init__(**kwargs)

        # Properties
        self.starred: Optional[bool] = starred

        # Edges
        self.label: EdgeList["Label"] = EdgeList("label", "Label", label)


class EmailMessage(Message):
    description = """A single email message."""
    properties = Message.properties + ["starred"]
    edges = Message.edges + ["bcc", "cc", "message", "replyTo"]

    def __init__(
        self,
        starred: bool = None,
        bcc: EdgeList["Account"] = None,
        cc: EdgeList["Account"] = None,
        message: EdgeList["EmailMessage"] = None,
        replyTo: EdgeList["Account"] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        # Properties
        self.starred: Optional[bool] = starred

        # Edges
        self.bcc: EdgeList["Account"] = EdgeList("bcc", "Account", bcc)
        self.cc: EdgeList["Account"] = EdgeList("cc", "Account", cc)
        self.message: EdgeList["EmailMessage"] = EdgeList("message", "EmailMessage", message)
        self.replyTo: EdgeList["Account"] = EdgeList("replyTo", "Account", replyTo)



import random
import string

from loguru import logger

from ._central_schema import *
from .dataset import Dataset
from .itembase import Edge, Item, ItemBase
from .photo import Photo


def get_constructor(_type, plugin_class=None, plugin_package=None, extra=None):
    import pymemri.integrator_registry

    from .photo import Photo

    if (
        _type == "Indexer"
        and plugin_class is not None
        and hasattr(pymemri.integrator_registry, plugin_class)
    ):
        return getattr(pymemri.integrator_registry, plugin_class)
    if plugin_class is not None and plugin_package is not None:
        try:
            mod = __import__(plugin_package, fromlist=[plugin_class])
            dynamic = {plugin_class: getattr(mod, plugin_class)}
        except Exception as e:
            logger.error(f"Could not import {plugin_class}.{plugin_package}")
    else:
        dynamic = dict()

    classes = z = {**globals(), **locals(), **extra, **dynamic}

    if _type in classes:
        if _type == "Indexer":
            constructor = classes[plugin_class]
        else:
            i_class = "I" + _type
            if i_class in classes:
                constructor = classes[i_class]
            else:
                constructor = classes[_type]
    else:
        raise TypeError(f"Could not initialize item, type {_type} not registered in PodClient")
    return constructor


class PluginRun(Item):
    description = """Information about a Plugin container being run by the Pod."""
    properties = Item.properties + [
        "containerId",
        "containerImage",
        "authUrl",
        "authImage",
        "pluginModule",
        "pluginName",
        "status",
        "targetItemId",
        "error",
        "config",
        "progress",
        "webserverPort",
    ]
    edges = Item.edges + ["plugin", "view", "account", "trigger"]

    def __init__(
        self,
        containerImage: str = None,
        authUrl: str = None,
        authImage: str = None,
        pluginModule: str = None,
        pluginName: str = None,
        status: str = None,
        targetItemId: str = None,
        error: str = None,
        config: str = None,
        progress: float = None,
        webserverPort: int = None,
        containerId: str = None,
        plugin: EdgeList[Plugin] = None,
        view: EdgeList[CVUStoredDefinition] = None,
        account: EdgeList[Account] = None,
        trigger: EdgeList[Edge] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        id_ = (
            "".join([random.choice(string.hexdigits) for i in range(32)])
            if targetItemId is None
            else targetItemId
        )
        self.id = id_

        # Properties
        self.containerImage: Optional[str] = containerImage
        self.authUrl: Optional[str] = authUrl
        self.authImage: Optional[str] = authImage
        self.pluginModule: Optional[str] = pluginModule
        self.pluginName: Optional[str] = pluginName
        self.status: Optional[str] = status
        self.targetItemId: Optional[str] = id_
        self.error: Optional[str] = error
        self.config: Optional[str] = config
        self.progress: Optional[float] = progress
        self.webserverPort: Optional[int] = webserverPort
        self.containerId: Optional[str] = containerId

        # Edges
        self.plugin: EdgeList[Plugin] = EdgeList("plugin", "Plugin", plugin)
        self.view: EdgeList[CVUStoredDefinition] = EdgeList("view", "CVUStoredDefinition", view)
        self.account: EdgeList[Account] = EdgeList("account", "Account", account)
        self.trigger: EdgeList[Trigger] = EdgeList("trigger", "Account", trigger)


class Trigger(Item):
    properties = Item.properties + ["action", "triggerOn", "pluginRunId", "filterCreatedAfter"]

    def __init__(
        self,
        name: str = None,
        version: str = None,
        action: str = None,
        triggerOn: str = None,
        pluginRunId: str = None,
        filterCreatedAfter: datetime = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Properties
        self.name: str = name
        self.version: str = version
        self.action: str = action
        self.triggerOn: str = triggerOn
        self.pluginRunId: str = pluginRunId
        self.filterCreatedAfter: datetime = filterCreatedAfter
